# Copyright © 2020 Interplanetary Database Association e.V.,
# BigchainDB and IPDB software contributors.
# SPDX-License-Identifier: (Apache-2.0 AND CC-BY-4.0)
# Code is Apache-2.0 and docs are CC-BY-4.0

__version__ = '2.2.4.0'
__short_version__ = '2.2'

# Supported Tendermint versions
__tm_supported_versions__ = ["0.31.5", "0.22.8"]
